// ---------------------------------------------------------------------
// exemple-timer-create-02.c
// Fichier d'exemple du livre "Solutions Temps-Reel sous Linux"
// (C) 2012 - Christophe BLAESS
// http://christophe.blaess.fr
// ---------------------------------------------------------------------

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>


static timer_t timer = (timer_t) 0;
static long int * mesure = NULL;
static int nb_mesures = 0;
static int nb_total_mesures = 0;

void handler_signal(int inutilise)
{
	struct timespec heure;
	static struct timespec precedente = { 0, 0} ;

	clock_gettime (CLOCK_REALTIME, & heure);
	if (precedente.tv_sec > 0) {
		mesure[nb_mesures]  = heure.tv_sec  - precedente.tv_sec;
		mesure[nb_mesures] *= 1000000000;
		mesure[nb_mesures] += heure.tv_nsec - precedente.tv_nsec;
		nb_mesures ++;
	}
	precedente = heure;
	if (nb_mesures == nb_total_mesures)
		timer_delete(timer);
}


int main(int argc, char * argv[])
{
	long int periode;
	struct sigevent event;
	struct itimerspec spec;
	
	if ((argc != 2) || (sscanf(argv[1], "%ld", & periode) != 1)){
		fprintf(stderr, "usage: %s periode_en_microsec\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	// Allouer le tableau de mesures
	nb_total_mesures = 5000000 / periode; // 5 secondes
	if ((mesure = calloc(nb_total_mesures, sizeof(struct timespec))) == NULL) {
		perror("calloc");
		exit(EXIT_FAILURE);
	}

	// Configurer le timer
	signal(SIGRTMIN, handler_signal);
	event.sigev_notify = SIGEV_SIGNAL;
	event.sigev_signo  = SIGRTMIN;
	periode = periode * 1000;  // en nanosec
	spec.it_interval.tv_sec  = periode / 1000000000;
	spec.it_interval.tv_nsec = periode % 1000000000;
	spec.it_value = spec.it_interval;
	// Allouer le timer
	if (timer_create(CLOCK_REALTIME, & event, & timer) != 0) {
		perror("timer_create");
		exit(EXIT_FAILURE);
	} 
	// Programmer le timer
	if (timer_settime(timer, 0, &spec, NULL) != 0) {
		perror("timer_settime");
		exit(EXIT_FAILURE);
	}
	// Attendre le dernier signal
	while (nb_mesures < nb_total_mesures)
		pause();
	// Afficher les mesures en micro-secondes
	for (nb_mesures = 0; nb_mesures < nb_total_mesures; nb_mesures ++)
		printf("%ld\n", mesure[nb_mesures]/1000);
	return EXIT_SUCCESS;
}

