// ---------------------------------------------------------------------
// exemple-threads-rr.c
// Fichier d'exemple du livre "Solutions Temps-Reel sous Linux"
// (C) 2012 - Christophe BLAESS
// http://christophe.blaess.fr
// ---------------------------------------------------------------------

#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#define NB_THREADS 4

void * fonction_thread(void * arg)
{
	int numero = (int) arg;

	struct timeval debut, precedente, heure;
	long long int difference;

	sleep(2);

	gettimeofday(& debut, NULL);

	heure = debut;

	while (1) {
		do {
			precedente = heure;
			gettimeofday(& heure, NULL);
			if ((heure.tv_sec - debut.tv_sec) > 10) // Maxi 10 secondes;
				pthread_exit(NULL);
			difference  = heure.tv_sec - precedente.tv_sec;
			difference *= 1000000;
			difference += heure.tv_usec - precedente.tv_usec;
		} while (difference < 1000); // Preemption d'au moins une milliseconde
		fprintf(stdout, "[%ld.%06ld] %d preempte\n",
		        precedente.tv_sec, precedente.tv_usec, numero);
		fprintf(stdout, "[%ld.%06ld] %d active\n",
		        heure.tv_sec, heure.tv_usec, numero);
	}
	
	return NULL;
}


int main(void)
{	
	int i;
	int err;
	pthread_attr_t attr;
	struct sched_param param;
	pthread_t thr[NB_THREADS];

	pthread_attr_init (& attr);
	if ((err = pthread_attr_setschedpolicy(& attr, SCHED_RR)) != 0) {
		fprintf(stderr, "setschedpolicy: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}
	if ((err = pthread_attr_setinheritsched(& attr, PTHREAD_EXPLICIT_SCHED)) != 0) {
		fprintf(stderr, "setinheritsched: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}
	param.sched_priority = 40;
	if ((err = pthread_attr_setschedparam(& attr, & param)) != 0) {
		fprintf(stderr, "setschedparam: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}
	for (i = 0; i < NB_THREADS; i ++) {
		if ((err = pthread_create(& (thr[i]), & attr, fonction_thread, (void *) (i+1))) != 0) {
			fprintf(stderr, "pthread_create: %s\n", strerror(err));
			exit(EXIT_FAILURE);
		}
	}
	for (i = 0; i < NB_THREADS; i ++)
		pthread_join(thr[i], NULL);

	return EXIT_SUCCESS;
}

