// ---------------------------------------------------------------------
// exemple-commutation-processus.c
// Fichier d'exemple du livre "Solutions Temps-Reel sous Linux"
// (C) 2012 - Christophe BLAESS
// http://christophe.blaess.fr
// ---------------------------------------------------------------------

#include <fcntl.h>
#include <sched.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/wait.h>

static sem_t * semaphore = NULL;

#define NB_COMMUTATIONS_MAX  1000
static int * nb_commutations = 0;
static long int * durees_commutations = NULL;

static struct timeval * avant_commutation = NULL;

void fils_haute_prio (void)
{
	struct timeval apres_commutation;
	usleep(100000); // 100 ms
	while ((*nb_commutations) < NB_COMMUTATIONS_MAX) {
		sem_wait(semaphore);
		gettimeofday(& apres_commutation, NULL);
		durees_commutations[(*nb_commutations)]  = apres_commutation.tv_sec - avant_commutation->tv_sec;
		durees_commutations[(*nb_commutations)] *= 1000000;
		durees_commutations[(*nb_commutations)] += apres_commutation.tv_usec - avant_commutation->tv_usec;
		(*nb_commutations) ++;
		sem_post(semaphore);
		usleep(1000); // 1 ms
	}
}


void fils_basse_prio (void)
{
	sem_wait(semaphore);
	usleep(200000); // 200 ms
	while ((*nb_commutations) < NB_COMMUTATIONS_MAX) {
		gettimeofday(avant_commutation, NULL);
		sem_post(semaphore);
		sem_wait(semaphore);
		usleep(1000); // 2 ms
	}
	sem_post(semaphore);
}

int main(void)
{	
	int i;
	int md;
	struct sched_param param;

	semaphore = sem_open("/essai-01", O_CREAT | O_RDWR, 0600, 1);
	if (semaphore == SEM_FAILED) {
		perror("/essai-01");
		exit(EXIT_FAILURE);
	}
	sem_unlink("/essai-01");

	md = shm_open("/essai-02", O_CREAT | O_RDWR, 0600);
	if (md < 0) {
		perror("/essai-02");
		exit(EXIT_FAILURE);
	}
	ftruncate(md, sizeof(struct timeval));
	avant_commutation = mmap(NULL, sizeof(struct timeval), PROT_READ | PROT_WRITE, MAP_SHARED, md, 0);
	if (avant_commutation == MAP_FAILED) {
		perror("/essai-02");
		exit(EXIT_FAILURE);
	}
	shm_unlink("/essai-02");

	md = shm_open("/essai-03", O_CREAT | O_RDWR, 0600);
	if (md < 0) {
		perror("/essai-03");
		exit(EXIT_FAILURE);
	}
	ftruncate(md, NB_COMMUTATIONS_MAX * sizeof(long int));
	durees_commutations = mmap(NULL, NB_COMMUTATIONS_MAX * sizeof(long int), PROT_READ | PROT_WRITE, MAP_SHARED, md, 0);
	if (durees_commutations == MAP_FAILED) {
		perror("/essai-03");
		exit(EXIT_FAILURE);
	}
	shm_unlink("/essai-03");
	
	md = shm_open("/essai-04", O_CREAT | O_RDWR, 0600);
	if (md < 0) {
		perror("/essai-04");
		exit(EXIT_FAILURE);
	}
	ftruncate(md, sizeof(int));
	nb_commutations = mmap(NULL, sizeof(int), PROT_READ | PROT_WRITE, MAP_SHARED, md, 0);
	if (nb_commutations == MAP_FAILED) {
		perror("/essai-04");
		exit(EXIT_FAILURE);
	}
	(* nb_commutations) = 0;
	shm_unlink("/essai-04");

	if (fork() == 0) {
		/* Premier fils */
		param.sched_priority = 20;	
		if (sched_setscheduler(0, SCHED_FIFO, & param) != 0) {
			perror("setscheduler");
			exit(EXIT_FAILURE);
		}
		fils_basse_prio();
		exit(EXIT_SUCCESS);
	}

	if (fork() == 0) {
		/* Second fils */
		param.sched_priority = 80;	
		if (sched_setscheduler(0, SCHED_FIFO, & param) != 0) {
			perror("setscheduler");
			exit(EXIT_FAILURE);
		}
		fils_haute_prio();
		exit(EXIT_SUCCESS);
	}
	waitpid(-1 , NULL, 0);
	waitpid(-1 , NULL, 0);

	for (i = 0; i < NB_COMMUTATIONS_MAX; i ++)
		printf("%ld\n", durees_commutations[i]);

	return EXIT_SUCCESS;
}

