// ---------------------------------------------------------------------
// exemple-commutation-threads.c
// Fichier d'exemple du livre "Solutions Temps-Reel sous Linux"
// (C) 2012 - Christophe BLAESS
// http://christophe.blaess.fr
// ---------------------------------------------------------------------

#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

static pthread_mutex_t  mutex  = PTHREAD_MUTEX_INITIALIZER;

#define NB_COMMUTATIONS_MAX  1000
static int nb_commutations = 0;
static long int durees_commutations[NB_COMMUTATIONS_MAX];

static struct timeval avant_commutation;


void * thread_haute_prio (void * unused)
{
	struct timeval apres_commutation;
	usleep(100000); // ms
	while (nb_commutations < NB_COMMUTATIONS_MAX) {
		pthread_mutex_lock(& mutex);
		gettimeofday(& apres_commutation, NULL);
		durees_commutations[nb_commutations]  = apres_commutation.tv_sec - avant_commutation.tv_sec;
		durees_commutations[nb_commutations] *= 1000000;
		durees_commutations[nb_commutations] += apres_commutation.tv_usec - avant_commutation.tv_usec;
		nb_commutations ++;
		pthread_mutex_unlock(& mutex);
		usleep(10000); // 10 ms
	}
	return NULL;
}


void * thread_basse_prio (void * unused)
{
	pthread_mutex_lock(& mutex);
	usleep(200000); // 200 ms
	while (nb_commutations < NB_COMMUTATIONS_MAX) {
		gettimeofday(& avant_commutation, NULL);
		pthread_mutex_unlock(& mutex);
		pthread_mutex_lock(& mutex);
		usleep(10000); // 10 ms
	}
	pthread_mutex_unlock(& mutex);
	return NULL;
}

int main(void)
{	
	int i;
	int err;
	pthread_attr_t attr;
	struct sched_param param;
	pthread_t thr[2];

	pthread_attr_init (& attr);
	if ((err = pthread_attr_setschedpolicy(& attr, SCHED_FIFO)) != 0) {
		fprintf(stderr, "setschedpolicy: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}
	if ((err = pthread_attr_setinheritsched(& attr, PTHREAD_EXPLICIT_SCHED)) != 0) {
		fprintf(stderr, "setinheritsched: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}

	param.sched_priority = 20;	
	if ((err = pthread_attr_setschedparam(& attr, & param)) != 0) {
		fprintf(stderr, "setschedparam: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}
	if ((err = pthread_create(& (thr[0]), & attr, thread_basse_prio, NULL)) != 0) {
		fprintf(stderr, "pthread_create: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}

	param.sched_priority = 80;
	if ((err = pthread_attr_setschedparam(& attr, & param)) != 0) {
		fprintf(stderr, "setschedparam: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}
	if ((err = pthread_create(& (thr[1]), & attr, thread_haute_prio, NULL)) != 0) {
		fprintf(stderr, "pthread_create: %s\n", strerror(err));
		exit(EXIT_FAILURE);
	}

	pthread_join(thr[0], NULL);
	pthread_join(thr[1], NULL);

	for (i = 0; i < NB_COMMUTATIONS_MAX; i ++)
		printf("%ld\n", durees_commutations[i]);

	return EXIT_SUCCESS;
}

